/**
 * Copyright (c) 2016-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 */

#include "IRCode.h"

#include <algorithm>
#include <boost/bimap/bimap.hpp>
#include <boost/bimap/unordered_set_of.hpp>
#include <boost/numeric/conversion/cast.hpp>
#include <memory>
#include <unordered_set>
#include <list>

#include "ControlFlow.h"
#include "Debug.h"
#include "DexClass.h"
#include "DexDebugInstruction.h"
#include "DexOpcode.h"
#include "DexUtil.h"
#include "IRInstruction.h"
#include "Transform.h"
#include "Util.h"

MethodItemEntry::MethodItemEntry(const MethodItemEntry& that)
    : type(that.type) {
  switch (type) {
  case MFLOW_TRY:
    tentry = that.tentry;
    break;
  case MFLOW_CATCH:
    centry = that.centry;
    break;
  case MFLOW_OPCODE:
    insn = that.insn;
    break;
  case MFLOW_DEX_OPCODE:
    dex_insn = that.dex_insn;
    break;
  case MFLOW_TARGET:
    target = that.target;
    break;
  case MFLOW_DEBUG:
    new (&dbgop) std::unique_ptr<DexDebugInstruction>(that.dbgop->clone());
    break;
  case MFLOW_POSITION:
    new (&pos) std::unique_ptr<DexPosition>(new DexPosition(*that.pos));
    break;
  case MFLOW_FALLTHROUGH:
    break;
  }
}

MethodItemEntry::~MethodItemEntry() {
  switch (type) {
    case MFLOW_TRY:
      delete tentry;
      break;
    case MFLOW_CATCH:
      delete centry;
      break;
    case MFLOW_TARGET:
      delete target;
      break;
    case MFLOW_DEBUG:
      dbgop.~unique_ptr<DexDebugInstruction>();
      break;
    case MFLOW_POSITION:
      pos.~unique_ptr<DexPosition>();
      break;
    case MFLOW_OPCODE:
    case MFLOW_DEX_OPCODE:
    case MFLOW_FALLTHROUGH:
      /* nothing to delete */
      break;
  }
}

void MethodItemEntry::gather_strings(std::vector<DexString*>& lstring) const {
  switch (type) {
  case MFLOW_TRY:
    break;
  case MFLOW_CATCH:
    break;
  case MFLOW_OPCODE:
    insn->gather_strings(lstring);
    break;
  case MFLOW_DEX_OPCODE:
    dex_insn->gather_strings(lstring);
    break;
  case MFLOW_TARGET:
    break;
  case MFLOW_DEBUG:
    dbgop->gather_strings(lstring);
    break;
  case MFLOW_POSITION:
    // although DexPosition contains strings, these strings don't find their
    // way into the APK
    break;
  case MFLOW_FALLTHROUGH:
    break;
  }
}

void MethodItemEntry::gather_methods(std::vector<DexMethodRef*>& lmethod) const {
  switch (type) {
  case MFLOW_TRY:
    break;
  case MFLOW_CATCH:
    break;
  case MFLOW_OPCODE:
    insn->gather_methods(lmethod);
    break;
  case MFLOW_DEX_OPCODE:
    dex_insn->gather_methods(lmethod);
    break;
  case MFLOW_TARGET:
    break;
  case MFLOW_DEBUG:
    dbgop->gather_methods(lmethod);
    break;
  case MFLOW_POSITION:
    break;
  case MFLOW_FALLTHROUGH:
    break;
  }
}

void MethodItemEntry::gather_fields(std::vector<DexFieldRef*>& lfield) const {
  switch (type) {
  case MFLOW_TRY:
    break;
  case MFLOW_CATCH:
    break;
  case MFLOW_OPCODE:
    insn->gather_fields(lfield);
    break;
  case MFLOW_DEX_OPCODE:
    dex_insn->gather_fields(lfield);
    break;
  case MFLOW_TARGET:
    break;
  case MFLOW_DEBUG:
    dbgop->gather_fields(lfield);
    break;
  case MFLOW_POSITION:
    break;
  case MFLOW_FALLTHROUGH:
    break;
  }
}

void MethodItemEntry::gather_types(std::vector<DexType*>& ltype) const {
  switch (type) {
  case MFLOW_TRY:
    break;
  case MFLOW_CATCH:
    if (centry->catch_type != nullptr) {
      ltype.push_back(centry->catch_type);
    }
    break;
  case MFLOW_OPCODE:
    insn->gather_types(ltype);
    break;
  case MFLOW_DEX_OPCODE:
    dex_insn->gather_types(ltype);
    break;
  case MFLOW_TARGET:
    break;
  case MFLOW_DEBUG:
    dbgop->gather_types(ltype);
    break;
  case MFLOW_POSITION:
    break;
  case MFLOW_FALLTHROUGH:
    break;
  }
}

opcode::Branchingness MethodItemEntry::branchingness() const {
  switch (type) {
  case MFLOW_OPCODE:
    return opcode::branchingness(insn->opcode());
  case MFLOW_DEX_OPCODE:
    return opcode::branchingness(dex_insn->opcode());
  default:
    return opcode::BRANCH_NONE;
  }
}

////////////////////////////////////////////////////////////////////////////////

IRCode::~IRCode() {
  m_fmethod->clear_and_dispose(FatMethodDisposer());
  delete m_fmethod;
}

bool IRCode::structural_equals(const IRCode& other) {
  auto it1 = this->begin();
  auto it2 = other.begin();

  for (; it1 != this->end() && it2 != other.end(); it1++, it2++) {
    if (it1->type != it2->type) {
      return false;
    }

    if (it1->type == MFLOW_OPCODE) {
      if (*it1->insn != *it2->insn) {
        return false;
      }
    } else if (it1->type == MFLOW_TARGET) {
      auto branch_target1 = static_cast<BranchTarget*>(it1->target);
      auto branch_target2 = static_cast<BranchTarget*>(it2->target);

      if (branch_target1->index != branch_target2->index) {
        return false;
      }
    }
  }

  return it1 == this->end() && it2 == other.end();
}

boost::sub_range<FatMethod> IRCode::get_param_instructions() const {
  auto params_end = std::find_if_not(
      m_fmethod->begin(), m_fmethod->end(), [&](const MethodItemEntry& mie) {
        return mie.type == MFLOW_FALLTHROUGH ||
               (mie.type == MFLOW_OPCODE &&
                opcode::is_load_param(mie.insn->opcode()));
      });
  return boost::sub_range<FatMethod>(m_fmethod->begin(), params_end);
}

void IRCode::gather_catch_types(std::vector<DexType*>& ltype) const {
  for (auto& mie : *m_fmethod) {
    if (mie.type != MFLOW_CATCH) {
      continue;
    }
    if (mie.centry->catch_type != nullptr) {
      ltype.push_back(mie.centry->catch_type);
    }
  }
}

void IRCode::gather_types(std::vector<DexType*>& ltype) const {
  for (auto& mie : *m_fmethod) {
    mie.gather_types(ltype);
  }
  if (m_dbg) m_dbg->gather_types(ltype);
}

void IRCode::gather_strings(std::vector<DexString*>& lstring) const {
  for (auto& mie : *m_fmethod) {
    mie.gather_strings(lstring);
  }
  if (m_dbg) m_dbg->gather_strings(lstring);
}

void IRCode::gather_fields(std::vector<DexFieldRef*>& lfield) const {
  for (auto& mie : *m_fmethod) {
    mie.gather_fields(lfield);
  }
}

void IRCode::gather_methods(std::vector<DexMethodRef*>& lmethod) const {
  for (auto& mie : *m_fmethod) {
    mie.gather_methods(lmethod);
  }
}

namespace {

int bytecount(int32_t v) {
  int bytecount = 4;
  if ((int32_t)((int8_t)(v & 0xff)) == v) {
    bytecount = 1;
  } else if ((int32_t)((int16_t)(v & 0xffff)) == v) {
    bytecount = 2;
  }
  return bytecount;
}

DexOpcode goto_for_offset(int32_t offset) {
  if (offset == 0) {
    return OPCODE_GOTO_32;
  }
  switch (bytecount(offset)) {
  case 1:
    return OPCODE_GOTO;
  case 2:
    return OPCODE_GOTO_16;
  case 4:
    return OPCODE_GOTO_32;
  default:
    always_assert_log(false, "Invalid bytecount %d", offset);
  }
}

namespace {

using namespace boost::bimaps;

struct Entry {};
struct Addr {};

typedef bimap<tagged<MethodItemEntry*, Entry>,
              unordered_set_of<tagged<uint32_t, Addr>>>
    EntryAddrBiMap;

} // namespace

static MethodItemEntry* get_target(
    const MethodItemEntry* mei,
    const EntryAddrBiMap& bm) {
  uint32_t base = bm.by<Entry>().at(const_cast<MethodItemEntry*>(mei));
  int offset = mei->dex_insn->offset();
  uint32_t target = base + offset;
  always_assert_log(
      bm.by<Addr>().count(target) != 0,
      "Invalid opcode target %08x[%p](%08x) %08x in get_target %s\n",
      base,
      mei,
      offset,
      target,
      SHOW(mei->insn));
  return bm.by<Addr>().at(target);
}

static void insert_branch_target(FatMethod* fm,
                                 MethodItemEntry* target,
                                 MethodItemEntry* src) {
  BranchTarget* bt = new BranchTarget();
  bt->type = BRANCH_SIMPLE;
  bt->src = src;

  fm->insert(fm->iterator_to(*target), *(new MethodItemEntry(bt)));
}

// Returns true if the offset could be encoded without modifying fm.
bool encode_offset(FatMethod* fm,
                   MethodItemEntry* branch_op_mie,
                   int32_t offset) {
  DexOpcode bop = branch_op_mie->dex_insn->opcode();
  if (is_goto(bop)) {
    DexOpcode goto_op = goto_for_offset(offset);
    if (goto_op != bop) {
      auto insn = branch_op_mie->dex_insn;
      branch_op_mie->dex_insn = new DexInstruction(goto_op);
      delete insn;
      return false;
    }
  } else if (is_conditional_branch(bop)) {
    // if-* opcodes can only encode up to 16-bit offsets. To handle larger ones
    // we use a goto/32 and have the inverted if-* opcode skip over it. E.g.
    //
    //   if-gt <large offset>
    //   nop
    //
    // becomes
    //
    //   if-le <label>
    //   goto/32 <large offset>
    //   label:
    //   nop
    if (bytecount(offset) > 2) {
      auto insn = branch_op_mie->dex_insn;
      branch_op_mie->dex_insn = new DexInstruction(OPCODE_GOTO_32);

      DexOpcode inverted = opcode::invert_conditional_branch(bop);
      MethodItemEntry* mei = new MethodItemEntry(new DexInstruction(inverted));
      mei->dex_insn->set_src(0, insn->src(0));
      fm->insert(fm->iterator_to(*branch_op_mie), *mei);

      // this iterator should always be valid -- an if-* instruction cannot
      // be the last opcode in a well-formed method
      auto next_insn_it = std::next(fm->iterator_to(*branch_op_mie));
      insert_branch_target(fm, &*next_insn_it, mei);

      delete insn;
      return false;
    }
  } else {
    always_assert_log(bop == OPCODE_FILL_ARRAY_DATA,
                      "Unexpected opcode %s",
                      SHOW(*branch_op_mie));
  }
  branch_op_mie->dex_insn->set_offset(offset);
  return true;
}

static bool multi_target_compare_index(const BranchTarget* a,
                                       const BranchTarget* b) {
  return (a->index < b->index);
}

static bool multi_contains_gaps(const std::vector<BranchTarget*>& targets) {
  int32_t key = targets.front()->index;
  for (auto target : targets) {
    if (target->index != key) return true;
    key++;
  }
  return false;
}

static void insert_multi_branch_target(FatMethod* fm,
                                       int32_t index,
                                       MethodItemEntry* target,
                                       MethodItemEntry* src) {
  BranchTarget* bt = new BranchTarget();
  bt->type = BRANCH_MULTI;
  bt->src = src;
  bt->index = index;

  fm->insert(fm->iterator_to(*target), *(new MethodItemEntry(bt)));
}

static int32_t read_int32(const uint16_t*& data) {
  int32_t result;
  memcpy(&result, data, sizeof(int32_t));
  data += 2;
  return result;
}

static void shard_multi_target(FatMethod* fm,
                               DexOpcodeData* fopcode,
                               MethodItemEntry* src,
                               const EntryAddrBiMap& bm) {
  const uint16_t* data = fopcode->data();
  uint16_t entries = *data++;
  auto ftype = fopcode->opcode();
  uint32_t base = bm.by<Entry>().at(src);
  if (ftype == FOPCODE_PACKED_SWITCH) {
    int32_t index = read_int32(data);
    for (int i = 0; i < entries; i++) {
      uint32_t targetaddr = base + read_int32(data);
      auto target = bm.by<Addr>().at(targetaddr);
      insert_multi_branch_target(fm, index, target, src);
      index++;
    }
  } else if (ftype == FOPCODE_SPARSE_SWITCH) {
    const uint16_t* tdata = data + 2 * entries;  // entries are 32b
    for (int i = 0; i < entries; i++) {
      int32_t index = read_int32(data);
      uint32_t targetaddr = base + read_int32(tdata);
      auto target = bm.by<Addr>().at(targetaddr);
      insert_multi_branch_target(fm, index, target, src);
    }
  } else {
    always_assert_log(false, "Bad fopcode 0x%04x in shard_multi_target", ftype);
  }
}

static void generate_branch_targets(
    FatMethod* fm,
    const EntryAddrBiMap& bm,
    const std::unordered_map<MethodItemEntry*, DexOpcodeData*>& entry_to_data) {
  for (auto miter = fm->begin(); miter != fm->end(); miter++) {
    MethodItemEntry* mentry = &*miter;
    if (mentry->type == MFLOW_DEX_OPCODE) {
      auto insn = mentry->dex_insn;
      if (is_branch(insn->opcode())) {
        if (is_multi_branch(insn->opcode())) {
          auto* fopcode_entry = get_target(mentry, bm);
          auto* fopcode = entry_to_data.at(fopcode_entry);
          shard_multi_target(fm, fopcode, mentry, bm);
          delete fopcode;
          // TODO: erase fopcode from map
        } else {
          auto target = get_target(mentry, bm);
          insert_branch_target(fm, target, mentry);
        }
      }
    }
  }
}

static void associate_debug_entries(FatMethod* fm,
                                    DexDebugItem& dbg,
                                    const EntryAddrBiMap& bm) {
  for (auto& entry : dbg.get_entries()) {
    auto insert_point = bm.by<Addr>().at(entry.addr);
    MethodItemEntry* mentry;
    switch (entry.type) {
      case DexDebugEntryType::Instruction:
        mentry = new MethodItemEntry(std::move(entry.insn));
        break;
      case DexDebugEntryType::Position:
        mentry = new MethodItemEntry(std::move(entry.pos));
        break;
    }
    fm->insert(fm->iterator_to(*insert_point), *mentry);
  }
  dbg.get_entries().clear();
}

static void associate_try_items(FatMethod* fm,
                                DexCode& code,
                                const EntryAddrBiMap& bm) {
  auto const& tries = code.get_tries();
  for (auto& tri : tries) {
    MethodItemEntry* catch_start = nullptr;
    CatchEntry* last_catch = nullptr;
    for (auto catz : tri->m_catches) {
      auto catzop = bm.by<Addr>().at(catz.second);
      TRACE(MTRANS, 3, "try_catch %08x mei %p\n", catz.second, catzop);
      auto catch_mei = new MethodItemEntry(catz.first);
      catch_start = catch_start == nullptr ? catch_mei : catch_start;
      if (last_catch != nullptr) {
        last_catch->next = catch_mei;
      }
      last_catch = catch_mei->centry;
      fm->insert(fm->iterator_to(*catzop), *catch_mei);
    }

    auto begin = bm.by<Addr>().at(tri->m_start_addr);
    TRACE(MTRANS, 3, "try_start %08x mei %p\n", tri->m_start_addr, begin);
    auto try_start = new MethodItemEntry(TRY_START, catch_start);
    fm->insert(fm->iterator_to(*begin), *try_start);
    uint32_t lastaddr = tri->m_start_addr + tri->m_insn_count;
    auto end = bm.by<Addr>().at(lastaddr);
    TRACE(MTRANS, 3, "try_end %08x mei %p\n", lastaddr, end);
    auto try_end = new MethodItemEntry(TRY_END, catch_start);
    fm->insert(fm->iterator_to(*end), *try_end);
  }
}

/*
 * Populate IRCode with load-param opcodes corresponding to the method
 * prototype. For example, a static method with proto "V(IJLfoo;)" and
 * no temp_regs will translate to
 *
 *   IOPCODE_LOAD_PARAM v0
 *   IOPCODE_LOAD_PARAM_WIDE v1
 *   IOPCODE_LOAD_PARAM_OBJECT v3
 */
void generate_load_params(const DexMethod* method,
                          size_t temp_regs,
                          IRCode* code) {
  auto args = method->get_proto()->get_args()->get_type_list();
  auto param_reg = temp_regs;
  if (!is_static(method)) {
    auto insn = new IRInstruction(IOPCODE_LOAD_PARAM_OBJECT);
    insn->set_dest(param_reg++);
    code->push_back(insn);
  }
  for (DexType* arg : args) {
    DexOpcode op;
    auto prev_reg = param_reg;
    if (is_wide_type(arg)) {
      param_reg += 2;
      op = IOPCODE_LOAD_PARAM_WIDE;
    } else {
      param_reg += 1;
      op = is_primitive(arg) ? IOPCODE_LOAD_PARAM : IOPCODE_LOAD_PARAM_OBJECT;
    }
    auto insn = new IRInstruction(op);
    insn->set_dest(prev_reg);
    code->push_back(insn);
  }
  code->set_registers_size(param_reg);
}

void translate_dex_to_ir(
    FatMethod* fmethod,
    const EntryAddrBiMap& bm,
    const std::unordered_map<MethodItemEntry*, DexOpcodeData*>& entry_to_data) {
  for (auto it = fmethod->begin(); it != fmethod->end(); ++it) {
    if (it->type != MFLOW_DEX_OPCODE) {
      continue;
    }
    auto* dex_insn = it->dex_insn;
    auto op = dex_insn->opcode();
    auto* insn = new IRInstruction(op);
    always_assert(!is_fopcode(op));

    IRInstruction* move_result_pseudo{nullptr};
    if (insn->dests_size()) {
      insn->set_dest(dex_insn->dest());
    } else if (opcode::may_throw(op)) {
      if (op == OPCODE_CHECK_CAST) {
        move_result_pseudo =
            new IRInstruction(IOPCODE_MOVE_RESULT_PSEUDO_OBJECT);
        move_result_pseudo->set_dest(dex_insn->src(0));
      } else if (dex_insn->dests_size()) {
        DexOpcode move_op;
        if (opcode_impl::dest_is_wide(op)) {
          move_op = IOPCODE_MOVE_RESULT_PSEUDO_WIDE;
        } else if (opcode_impl::dest_is_object(op)) {
          move_op = IOPCODE_MOVE_RESULT_PSEUDO_OBJECT;
        } else {
          move_op = IOPCODE_MOVE_RESULT_PSEUDO;
        }
        move_result_pseudo = new IRInstruction(move_op);
        move_result_pseudo->set_dest(dex_insn->dest());
      }
    }

    insn->set_arg_word_count(dex_insn->srcs_size()); // XXX: should we have a better API?
    for (size_t i = 0; i < dex_insn->srcs_size(); ++i) {
      insn->set_src(i, dex_insn->src(i));
    }
    if (opcode::dest_is_src(op)) {
      insn->set_opcode(convert_2to3addr(op));
    }
    if (opcode::has_literal(op)) {
      insn->set_literal(dex_insn->get_literal());
    }
    if (opcode::has_range(op)) {
      insn->set_range_base(dex_insn->range_base());
      insn->set_range_size(dex_insn->range_size());
    }
    if (dex_insn->has_string()) {
      insn->set_string(
          static_cast<const DexOpcodeString*>(dex_insn)->get_string());
    } else if (dex_insn->has_type()) {
      insn->set_type(static_cast<const DexOpcodeType*>(dex_insn)->get_type());
    } else if (dex_insn->has_field()) {
      insn->set_field(static_cast<const DexOpcodeField*>(dex_insn)->get_field());
    } else if (dex_insn->has_method()) {
      insn->set_method(
          static_cast<const DexOpcodeMethod*>(dex_insn)->get_method());
    } else if (op == OPCODE_FILL_ARRAY_DATA) {
      insn->set_data(entry_to_data.at(get_target(&*it, bm)));
    }

    it->type = MFLOW_OPCODE;
    it->insn = insn;
    if (move_result_pseudo != nullptr) {
      it = fmethod->insert(std::next(it),
                           *(new MethodItemEntry(move_result_pseudo)));
    }
  }
}

void balloon(DexMethod* method, FatMethod* fmethod) {
  auto dex_code = method->get_dex_code();
  auto instructions = dex_code->release_instructions();
  // This is a 1-to-1 map between MethodItemEntries of type MFLOW_OPCODE and
  // address offsets.
  EntryAddrBiMap bm;
  std::unordered_map<MethodItemEntry*, DexOpcodeData*> entry_to_data;

  uint32_t addr = 0;
  for (auto insn : *instructions) {
    MethodItemEntry* mei;
    if (insn->opcode() == OPCODE_NOP || is_fopcode(insn->opcode())) {
      // We have to insert dummy entries for these opcodes so that try items
      // and debug entries that are adjacent to them can find the right
      // address.
      mei = new MethodItemEntry();
      if (is_fopcode(insn->opcode())) {
        entry_to_data.emplace(mei, static_cast<DexOpcodeData*>(insn));
      }
    } else {
      mei = new MethodItemEntry(insn);
    }
    fmethod->push_back(*mei);
    bm.insert(EntryAddrBiMap::relation(mei, addr));
    TRACE(MTRANS, 5, "%08x: %s[mei %p]\n", addr, SHOW(insn), mei);
    addr += insn->size();
  }
  bm.insert(EntryAddrBiMap::relation(&*fmethod->end(), addr));

  generate_branch_targets(fmethod, bm, entry_to_data);
  associate_try_items(fmethod, *dex_code, bm);
  translate_dex_to_ir(fmethod, bm, entry_to_data);
  auto debugitem = dex_code->get_debug_item();
  if (debugitem) {
    associate_debug_entries(fmethod, *debugitem, bm);
  }
}

/**
 * Map the `DexPositions` to a newly created clone. At the same time, it
 * preserves the relationship between a position and it's parent.
 */
std::unordered_map<DexPosition*, std::unique_ptr<DexPosition>>
get_old_to_new_position_copies(FatMethod* fmethod) {
  std::unordered_map<DexPosition*, std::unique_ptr<DexPosition>>
      old_position_to_new;
  for (auto& mie : *fmethod) {
    if (mie.type == MFLOW_POSITION) {
      old_position_to_new.emplace(mie.pos.get(),
                                  std::make_unique<DexPosition>(*mie.pos));
    }
  }

  for (auto& old_to_new : old_position_to_new) {
    DexPosition* old_pos = old_to_new.first;
    auto new_pos = old_to_new.second.get();

    // There may be dangling pointers to parent positions that have been deleted
    // So, we can't use the [] operator here because it would add
    // nullptr as a value in the map which would cause a segfault on a later
    // iteration of the loop. `.at()` is also not an option for the same reason.
    new_pos->parent = nullptr;
    if (old_pos->parent != nullptr) {
      auto search = old_position_to_new.find(old_pos->parent);
      if (search != old_position_to_new.end()) {
        new_pos->parent = search->second.get();
      }
    }
  }

  return old_position_to_new;
}

// TODO: merge this and MethodSplicer.
FatMethod* deep_copy_fmethod(FatMethod* old_fmethod) {
  FatMethod* fmethod = new FatMethod();

  std::unordered_map<DexPosition*, std::unique_ptr<DexPosition>>
      old_position_to_new = get_old_to_new_position_copies(old_fmethod);

  // Create a clone for each of the entries
  // and a mapping from old pointers to new pointers.
  std::unordered_map<MethodItemEntry*, MethodItemEntry*> old_mentry_to_new;
  for (auto& mie : *old_fmethod) {
    MethodItemEntry* copy_mie = new MethodItemEntry();
    copy_mie->type = mie.type;
    old_mentry_to_new[&mie] = copy_mie;
  }

  // now fill the fields of the `copy_mie`s
  for (auto& mie : *old_fmethod) {

    auto copy_mie = old_mentry_to_new.at(&mie);
    switch (mie.type) {
    case MFLOW_TRY:
      copy_mie->tentry =
          new TryEntry(mie.tentry->type,
                       mie.tentry->catch_start
                           ? old_mentry_to_new[mie.tentry->catch_start]
                           : nullptr);
      break;
    case MFLOW_CATCH:
      copy_mie->centry = new CatchEntry(mie.centry->catch_type);
      copy_mie->centry->next =
          mie.centry->next ? old_mentry_to_new[mie.centry->next] : nullptr;
      break;
    case MFLOW_TARGET:
      copy_mie->target = new BranchTarget();
      copy_mie->target->type = mie.target->type;
      copy_mie->target->index = mie.target->index;
      copy_mie->target->src = old_mentry_to_new[mie.target->src];
      break;
    case MFLOW_OPCODE:
      copy_mie->insn = new IRInstruction(*mie.insn);
      break;
    case MFLOW_DEBUG:
      new (&copy_mie->dbgop)
          std::unique_ptr<DexDebugInstruction>(mie.dbgop->clone());
      break;
    case MFLOW_POSITION:
      copy_mie->pos = std::move(old_position_to_new[mie.pos.get()]);
      break;
    case MFLOW_FALLTHROUGH:
      break;
    case MFLOW_DEX_OPCODE:
      always_assert_log(false, "DexInstruction not expected here!");
    default:
      not_reached();
    }

    fmethod->push_back(*copy_mie);
  }

  return fmethod;
}

} // namespace

IRCode::IRCode(): m_fmethod(new FatMethod()) {}

IRCode::IRCode(DexMethod* method): m_fmethod(new FatMethod()) {
  auto* dc = method->get_dex_code();
  generate_load_params(
      method, dc->get_registers_size() - dc->get_ins_size(), this);
  balloon(const_cast<DexMethod*>(method), m_fmethod);
  m_dbg = dc->release_debug_item();
}

IRCode::IRCode(DexMethod* method, size_t temp_regs)
    : m_fmethod(new FatMethod()) {
  always_assert(method->get_dex_code() == nullptr);
  generate_load_params(method, temp_regs, this);
}

IRCode::IRCode(const IRCode& code) {
  FatMethod* old_fmethod = code.m_fmethod;
  m_fmethod = deep_copy_fmethod(old_fmethod);
  m_registers_size = code.m_registers_size;
  if (code.m_dbg) {
    m_dbg = std::make_unique<DexDebugItem>(*code.m_dbg);
  }
}

void IRCode::remove_branch_targets(IRInstruction *branch_inst) {
  always_assert_log(is_branch(branch_inst->opcode()),
                    "Instruction is not a branch instruction.");
  for (auto miter = m_fmethod->begin(); miter != m_fmethod->end(); miter++) {
    MethodItemEntry* mentry = &*miter;
    if (mentry->type == MFLOW_TARGET) {
      BranchTarget* bt = mentry->target;
      auto btmei = bt->src;
      if (btmei->insn == branch_inst) {
        mentry->type = MFLOW_FALLTHROUGH;
        delete mentry->target;
      }
    }
  }
}

void IRCode::replace_branch(IRInstruction* from, IRInstruction* to) {
  always_assert(is_branch(from->opcode()));
  always_assert(is_branch(to->opcode()));
  for (auto& mentry : *m_fmethod) {
    if (mentry.type == MFLOW_OPCODE && mentry.insn == from) {
      mentry.insn = to;
      delete from;
      return;
    }
  }
  always_assert_log(
      false,
      "No match found while replacing '%s' with '%s'",
      SHOW(from),
      SHOW(to));
}

void IRCode::remove_debug_line_info(Block* block) {
  for (MethodItemEntry& mie : *block) {
    if (mie.type == MFLOW_POSITION) {
      mie.pos.release();
      mie.type = MFLOW_FALLTHROUGH;
    }
  }
}

void IRCode::replace_opcode_with_infinite_loop(IRInstruction* from) {
  IRInstruction* to = new IRInstruction(OPCODE_GOTO_32);
  auto miter = m_fmethod->begin();
  for (; miter != m_fmethod->end(); miter++) {
    MethodItemEntry* mentry = &*miter;
    if (mentry->type == MFLOW_OPCODE && mentry->insn == from) {
      if (is_branch(from->opcode())) {
        remove_branch_targets(from);
      }
      mentry->insn = to;
      delete from;
      break;
    }
  }
  always_assert_log(
      miter != m_fmethod->end(),
      "No match found while replacing '%s' with '%s'",
      SHOW(from),
      SHOW(to));
  auto target = new BranchTarget();
  target->type = BRANCH_SIMPLE;
  target->src = &*miter;
  m_fmethod->insert(miter, *(new MethodItemEntry(target)));
}

void IRCode::replace_opcode(IRInstruction* to_delete,
                            std::vector<IRInstruction*> replacements) {
  auto it = m_fmethod->begin();
  for (; it != m_fmethod->end(); it++) {
    if (it->type == MFLOW_OPCODE && it->insn == to_delete) {
      break;
    }
  }
  always_assert_log(it != m_fmethod->end(),
                    "No match found while replacing '%s'",
                    SHOW(to_delete));
  for (auto insn : replacements) {
    insert_before(it, insn);
  }
  remove_opcode(it);
}

void IRCode::replace_opcode(IRInstruction* from, IRInstruction* to) {
  always_assert_log(!is_branch(to->opcode()),
                    "You may want replace_branch instead");
  replace_opcode(from, std::vector<IRInstruction*>{to});
}

void IRCode::insert_after(IRInstruction* position,
                          const std::vector<IRInstruction*>& opcodes) {
  /* The nullptr case handling is strange-ish..., this will not work as expected
   *if
   * a method has a branch target as it's first instruction.
   *
   * To handle this case sanely, we'd need to export a interface based on
   * MEI's probably.
   *
   */
  for (auto const& mei : *m_fmethod) {
    if (mei.type == MFLOW_OPCODE &&
        (position == nullptr || mei.insn == position)) {
      auto insert_at = m_fmethod->iterator_to(mei);
      if (position != nullptr) {
        insert_at++;
        if (position->has_move_result_pseudo()) {
          insert_at++;
        }
      }
      for (auto* opcode : opcodes) {
        MethodItemEntry* mentry = new MethodItemEntry(opcode);
        m_fmethod->insert(insert_at, *mentry);
      }
      return;
    }
  }
  always_assert_log(false, "No match found");
}

FatMethod::iterator IRCode::insert_before(
    const FatMethod::iterator& position, MethodItemEntry& mie) {
  return m_fmethod->insert(position, mie);
}

FatMethod::iterator IRCode::insert_after(
    const FatMethod::iterator& position, MethodItemEntry& mie) {
  always_assert(position != m_fmethod->end());
  return m_fmethod->insert(std::next(position), mie);
}

/*
 * Param `insn` should be part of a switch...case statement. Find the case
 * block it is contained within and remove it. Then decrement the index of
 * all the other case blocks that are larger than the index of the removed
 * block so that the case numbers don't have any gaps and the switch can
 * still be encoded as a packed-switch opcode.
 *
 * We do the removal by removing the MFLOW_TARGET corresponding to that
 * case label. Its contents are dead code which will be removed by LocalDCE
 * later. (We could do it here too, but LocalDCE already knows how to find
 * block boundaries.)
 */
void IRCode::remove_switch_case(IRInstruction* insn) {

  TRACE(MTRANS, 3, "Removing switch case from: %s\n", SHOW(m_fmethod));
  // Check if we are inside switch method.
  const MethodItemEntry* switch_mei {nullptr};
  for (const auto& mei : InstructionIterable(this)) {
    auto op = mei.insn->opcode();
    if (opcode::is_load_param(op)) {
      continue;
    }
    assert_log(is_multi_branch(op), " Method is not a switch");
    switch_mei = &mei;
    break;
  }
  always_assert(switch_mei != nullptr);

  int target_count = 0;
  for (auto& mei : *m_fmethod) {
    if (mei.type == MFLOW_TARGET && mei.target->type == BRANCH_MULTI) {
      target_count++;
    }
  }
  assert_log(target_count != 0, " There should be atleast one target");
  if (target_count == 1) {
    auto excpt_str = DexString::make_string("Redex switch Exception");
    std::vector<IRInstruction*> excpt_block;
    create_runtime_exception_block(excpt_str, excpt_block);
    insert_after(insn, excpt_block);
    remove_opcode(insn);
    return;
  }

  // Find the starting MULTI Target point to delete.
  MethodItemEntry* target_mei = nullptr;
  for (auto miter = m_fmethod->begin(); miter != m_fmethod->end(); miter++) {
    MethodItemEntry* mentry = &*miter;
    if (mentry->type == MFLOW_TARGET) {
      target_mei = mentry;
    }
    // Check if insn belongs to the current block.
    if (mentry->type == MFLOW_OPCODE && mentry->insn == insn) {
      break;
    }
  }
  always_assert_log(target_mei != nullptr,
                    "Could not find target for %s in %s",
                    SHOW(insn),
                    SHOW(m_fmethod));

  for (const auto& mie : *m_fmethod) {
    if (mie.type == MFLOW_TARGET) {
      BranchTarget* bt = mie.target;
      if (bt->src == switch_mei && bt->index > target_mei->target->index) {
        bt->index -= 1;
      }
    }
  }

  target_mei->type = MFLOW_FALLTHROUGH;
  delete target_mei->target;
}

size_t IRCode::count_opcodes() const {
  size_t count {0};
  for (const auto& mie : *m_fmethod) {
    if (mie.type == MFLOW_OPCODE && !opcode::is_internal(mie.insn->opcode())) {
      ++count;
    }
  }
  return count;
}

size_t IRCode::sum_opcode_sizes() const {
  size_t size {0};
  for (const auto& mie : *m_fmethod) {
    if (mie.type == MFLOW_OPCODE) {
      size += mie.insn->size();
    }
  }
  return size;
}

void IRCode::remove_opcode(const FatMethod::iterator& it) {
  always_assert(it->type == MFLOW_OPCODE);
  auto insn = it->insn;
  always_assert(!opcode::is_move_result_pseudo(insn->opcode()));
  if (insn->has_move_result_pseudo()) {
    auto move_it = std::next(it);
    always_assert_log(
        move_it->type == MFLOW_OPCODE &&
            opcode::is_move_result_pseudo(move_it->insn->opcode()),
        "No move-result-pseudo found for %s",
        SHOW(insn));
    delete move_it->insn;
    move_it->type = MFLOW_FALLTHROUGH;
    move_it->insn = nullptr;
  }
  if (is_branch(insn->opcode())) {
    remove_branch_targets(insn);
  }
  it->type = MFLOW_FALLTHROUGH;
  it->insn = nullptr;
  delete insn;
}

void IRCode::remove_opcode(IRInstruction* insn) {
  for (auto& mei : *m_fmethod) {
    if (mei.type == MFLOW_OPCODE && mei.insn == insn) {
      auto it = m_fmethod->iterator_to(mei);
      remove_opcode(it);
      return;
    }
  }
  always_assert_log(false,
                    "No match found while removing '%s' from method",
                    SHOW(insn));
}

FatMethod::iterator IRCode::main_block() {
  return std::prev(get_param_instructions().end());
}

FatMethod::iterator IRCode::insert(FatMethod::iterator cur,
                                   IRInstruction* insn) {
  MethodItemEntry* mentry = new MethodItemEntry(insn);
  return m_fmethod->insert(cur, *mentry);
}

FatMethod::iterator IRCode::make_if_block(
    FatMethod::iterator cur,
    IRInstruction* insn,
    FatMethod::iterator* false_block) {
  auto if_entry = new MethodItemEntry(insn);
  *false_block = m_fmethod->insert(cur, *if_entry);
  auto bt = new BranchTarget();
  bt->src = if_entry;
  bt->type = BRANCH_SIMPLE;
  auto bentry = new MethodItemEntry(bt);
  return m_fmethod->insert(m_fmethod->end(), *bentry);
}

FatMethod::iterator IRCode::make_if_else_block(
    FatMethod::iterator cur,
    IRInstruction* insn,
    FatMethod::iterator* false_block,
    FatMethod::iterator* true_block) {
  // if block
  auto if_entry = new MethodItemEntry(insn);
  *false_block = m_fmethod->insert(cur, *if_entry);

  // end of else goto
  auto goto_entry = new MethodItemEntry(new IRInstruction(OPCODE_GOTO));
  auto goto_it = m_fmethod->insert(m_fmethod->end(), *goto_entry);

  // main block
  auto main_bt = new BranchTarget();
  main_bt->src = goto_entry;
  main_bt->type = BRANCH_SIMPLE;
  auto mb_entry = new MethodItemEntry(main_bt);
  auto main_block = m_fmethod->insert(goto_it, *mb_entry);

  // else block
  auto else_bt = new BranchTarget();
  else_bt->src = if_entry;
  else_bt->type = BRANCH_SIMPLE;
  auto eb_entry = new MethodItemEntry(else_bt);
  *true_block = m_fmethod->insert(goto_it, *eb_entry);

  return main_block;
}

FatMethod::iterator IRCode::make_switch_block(
    FatMethod::iterator cur,
    IRInstruction* insn,
    FatMethod::iterator* default_block,
    std::map<SwitchIndices, FatMethod::iterator>& cases) {
  auto switch_entry = new MethodItemEntry(insn);
  *default_block = m_fmethod->insert(cur, *switch_entry);
  FatMethod::iterator main_block = *default_block;
  for (auto case_it = cases.begin(); case_it != cases.end(); ++case_it) {
    auto goto_entry = new MethodItemEntry(new IRInstruction(OPCODE_GOTO));
    auto goto_it = m_fmethod->insert(m_fmethod->end(), *goto_entry);

    auto main_bt = new BranchTarget();
    main_bt->src = goto_entry;
    main_bt->type = BRANCH_SIMPLE;
    auto mb_entry = new MethodItemEntry(main_bt);
    main_block = m_fmethod->insert(++main_block, *mb_entry);

    // Insert all the branch targets jumping from the switch entry.
    // Keep updating the iterator of the case block to point right before the
    // GOTO going back to the end of the switch.
    for (auto idx : case_it->first) {
      auto case_bt = new BranchTarget();
      case_bt->src = switch_entry;
      case_bt->index = idx;
      case_bt->type = BRANCH_MULTI;
      auto eb_entry = new MethodItemEntry(case_bt);
      case_it->second = m_fmethod->insert(goto_it, *eb_entry);
    }
  }
  return main_block;
}

namespace {

using RegMap = transform::RegMap;

const char* DEBUG_ONLY show_reg_map(RegMap& map) {
  for (auto pair : map) {
    TRACE(INL, 5, "%u -> %u\n", pair.first, pair.second);
  }
  return "";
}

} // namespace

void IRCode::build_cfg(bool editable) {
  clear_cfg();
  m_cfg = std::make_unique<ControlFlowGraph>(m_fmethod, editable);
}

void IRCode::clear_cfg() {
  if (m_cfg && m_cfg->editable()) {
    m_fmethod = m_cfg->linearize();
  }

  m_cfg.reset();
  std::vector<FatMethod::iterator> fallthroughs;
  for (auto it = m_fmethod->begin(); it != m_fmethod->end(); ++it) {
    if (it->type == MFLOW_FALLTHROUGH) {
      fallthroughs.emplace_back(it);
    }
  }
  for (auto it : fallthroughs) {
    m_fmethod->erase_and_dispose(it, FatMethodDisposer());
  }
}

namespace ir_code_impl {

static uint16_t calc_outs_size(const IRCode* code) {
  uint16_t size {0};
  for (auto& mie : *code) {
    if (mie.type != MFLOW_DEX_OPCODE) {
      continue;
    }
    auto insn = mie.dex_insn;
    if (is_invoke_range(insn->opcode())) {
      size = std::max(size, boost::numeric_cast<uint16_t>(insn->range_size()));
    } else if (is_invoke(insn->opcode())) {
      size = std::max(size, boost::numeric_cast<uint16_t>(insn->srcs_size()));
    }
  }
  return size;
}

static void calculate_ins_size(const DexMethod* method,
                               DexCode* dex_code) {
  auto& args_list = method->get_proto()->get_args()->get_type_list();
  uint16_t ins_size{0};
  if (!is_static(method)) {
    ++ins_size;
  }
  for (auto arg : args_list) {
    if (is_wide_type(arg)) {
      ins_size += 2;
    } else {
      ++ins_size;
    }
  }
  dex_code->set_ins_size(ins_size);
}

} // namespace ir_code_impl

using namespace ir_code_impl;

std::unique_ptr<DexCode> IRCode::sync(const DexMethod* method) {
  auto dex_code = std::make_unique<DexCode>();
  try {
    calculate_ins_size(method, &*dex_code);
    dex_code->set_registers_size(m_registers_size);
    dex_code->set_outs_size(calc_outs_size(this));
    dex_code->set_debug_item(std::move(m_dbg));
    while (try_sync(dex_code.get()) == false)
      ;
  } catch (std::exception&) {
    fprintf(stderr, "Failed to sync %s\n%s\n", SHOW(method), SHOW(this));
    throw;
  }
  return dex_code;
}

bool IRCode::try_sync(DexCode* code) {
  std::unordered_map<MethodItemEntry*, uint32_t> entry_to_addr;
  uint32_t addr = 0;
  // Step 1, regenerate opcode list for the method, and
  // and calculate the opcode entries address offsets.
  TRACE(MTRANS, 5, "Emitting opcodes\n");
  for (auto miter = m_fmethod->begin(); miter != m_fmethod->end(); miter++) {
    MethodItemEntry* mentry = &*miter;
    TRACE(MTRANS, 5, "Analyzing mentry %p\n", mentry);
    entry_to_addr[mentry] = addr;
    if (mentry->type == MFLOW_DEX_OPCODE) {
      TRACE(MTRANS, 5, "Emitting mentry %p at %08x\n", mentry, addr);
      addr += mentry->dex_insn->size();
    }
  }
  // Step 2, Branch relaxation: calculate branch offsets for if-* and goto
  // opcodes, resizing them where necessary. Since resizing opcodes affects
  // address offsets, we need to iterate this to a fixed point.
  //
  // For instructions that use address offsets but never need resizing (i.e.
  // switch and fill-array-data opcodes), we calculate their offsets after
  // we have reached the fixed point.
  TRACE(MTRANS, 5, "Recalculating branches\n");
  std::vector<MethodItemEntry*> multi_branches;
  std::unordered_map<MethodItemEntry*, std::vector<BranchTarget*>> multis;
  std::unordered_map<BranchTarget*, uint32_t> multi_targets;
  bool needs_resync = false;
  for (auto miter = m_fmethod->begin(); miter != m_fmethod->end(); miter++) {
    MethodItemEntry* mentry = &*miter;
    if (entry_to_addr.find(mentry) == entry_to_addr.end()) {
      continue;
    }
    if (mentry->type == MFLOW_DEX_OPCODE) {
      auto opcode = mentry->dex_insn->opcode();
      if (is_multi_branch(opcode)) {
        multi_branches.push_back(mentry);
      }
    }
    if (mentry->type == MFLOW_TARGET) {
      BranchTarget* bt = mentry->target;
      if (bt->type == BRANCH_MULTI) {
        multis[bt->src].push_back(bt);
        multi_targets[bt] = entry_to_addr.at(mentry);
        // We can't fix the primary switch opcodes address until we emit
        // the fopcode, which comes later.
      } else if (bt->type == BRANCH_SIMPLE &&
                 is_branch(bt->src->dex_insn->opcode())) {
        MethodItemEntry* branch_op_mie = bt->src;
        auto branch_addr = entry_to_addr.find(branch_op_mie);
        always_assert_log(branch_addr != entry_to_addr.end(),
                          "%s refers to nonexistent branch instruction",
                          SHOW(*mentry));
        int32_t branch_offset = entry_to_addr.at(mentry) - branch_addr->second;
        if (branch_offset == 1) {
          needs_resync = true;
          branch_op_mie->type = MFLOW_FALLTHROUGH;
          mentry->type = MFLOW_FALLTHROUGH;
        } else {
          needs_resync |=
              !encode_offset(m_fmethod, branch_op_mie, branch_offset);
        }
      }
    }
  }
  if (needs_resync) {
    return false;
  }

  size_t num_align_nops{0};
  auto& opout = code->reset_instructions();
  for (auto& mie : *m_fmethod) {
    // We are assuming that fill-array-data-payload opcodes are always at
    // the end of the opcode stream (we enforce that during instruction
    // lowering). I.e. they are only followed by other fill-array-data-payload
    // opcodes. So adjusting their addresses here does not require re-running
    // branch relaxation.
    entry_to_addr.at(&mie) += num_align_nops;
    if (mie.type == MFLOW_TARGET &&
        mie.target->src->dex_insn->opcode() == OPCODE_FILL_ARRAY_DATA) {
      // This MFLOW_TARGET is right before a fill-array-data-payload opcode,
      // so we should make sure its address is aligned
      if (entry_to_addr.at(&mie) & 1) {
        opout.push_back(new DexInstruction(OPCODE_NOP));
        ++entry_to_addr.at(&mie);
        ++num_align_nops;
      }
      mie.target->src->dex_insn->set_offset(entry_to_addr.at(&mie) -
                                            entry_to_addr.at(mie.target->src));
      continue;
    }
    if (mie.type != MFLOW_DEX_OPCODE) {
      continue;
    }
    TRACE(MTRANS, 6, "Emitting insn %s\n", SHOW(mie.dex_insn));
    opout.push_back(mie.dex_insn);
  }
  addr += num_align_nops;

  TRACE(MTRANS, 5, "Emitting multi-branches\n");
  // Step 3, generate multi-branch fopcodes
  for (auto multiopcode : multi_branches) {
    auto& targets = multis[multiopcode];
    auto multi_insn = multiopcode->dex_insn;
    std::sort(targets.begin(), targets.end(), multi_target_compare_index);
    always_assert_log(!targets.empty(), "need to have targets");
    if (multi_contains_gaps(targets)) {
      // Emit sparse.
      const size_t count = (targets.size() * 4) + 2;
      auto sparse_payload = std::make_unique<uint16_t[]>(count);
      sparse_payload[0] = FOPCODE_SPARSE_SWITCH;
      sparse_payload[1] = targets.size();
      uint32_t* spkeys = (uint32_t*)&sparse_payload[2];
      uint32_t* sptargets =
          (uint32_t*)&sparse_payload[2 + (targets.size() * 2)];
      for (BranchTarget* target : targets) {
        *spkeys++ = target->index;
        *sptargets++ = multi_targets[target] - entry_to_addr.at(multiopcode);
      }
      // Emit align nop
      if (addr & 1) {
        DexInstruction* nop = new DexInstruction(OPCODE_NOP);
        opout.push_back(nop);
        addr++;
      }
      // Insert the new fopcode...
      DexInstruction* fop =
          new DexOpcodeData(sparse_payload.get(), (int)(count - 1));
      opout.push_back(fop);
      // re-write the source opcode with the address of the
      // fopcode, increment the address of the fopcode.
      multi_insn->set_offset(addr - entry_to_addr.at(multiopcode));
      multi_insn->set_opcode(OPCODE_SPARSE_SWITCH);
      addr += count;
    } else {
      // Emit packed.
      const size_t count = (targets.size() * 2) + 4;
      auto packed_payload = std::make_unique<uint16_t[]>(count);
      packed_payload[0] = FOPCODE_PACKED_SWITCH;
      packed_payload[1] = targets.size();
      uint32_t* psdata = (uint32_t*)&packed_payload[2];
      *psdata++ = targets.front()->index;
      for (BranchTarget* target : targets) {
        *psdata++ = multi_targets[target] - entry_to_addr.at(multiopcode);
      }
      // Emit align nop
      if (addr & 1) {
        DexInstruction* nop = new DexInstruction(0);
        opout.push_back(nop);
        addr++;
      }
      // Insert the new fopcode...
      DexInstruction* fop =
          new DexOpcodeData(packed_payload.get(), (int)(count - 1));
      opout.push_back(fop);
      // re-write the source opcode with the address of the
      // fopcode, increment the address of the fopcode.
      multi_insn->set_offset(addr - entry_to_addr.at(multiopcode));
      multi_insn->set_opcode(OPCODE_PACKED_SWITCH);
      addr += count;
    }
  }

  // Step 4, emit debug opcodes
  TRACE(MTRANS, 5, "Emitting debug opcodes\n");
  auto debugitem = code->get_debug_item();
  if (debugitem) {
    auto& entries = debugitem->get_entries();
    for (auto& mentry : *m_fmethod) {
      if (mentry.type == MFLOW_DEBUG) {
        entries.emplace_back(entry_to_addr.at(&mentry), std::move(mentry.dbgop));
      } else if (mentry.type == MFLOW_POSITION) {
        entries.emplace_back(entry_to_addr.at(&mentry), std::move(mentry.pos));
      }
    }
  }
  // Step 5, try/catch blocks
  TRACE(MTRANS, 5, "Emitting try items & catch handlers\n");
  auto& tries = code->get_tries();
  tries.clear();
  MethodItemEntry* active_try = nullptr;
  for (auto& mentry : *m_fmethod) {
    if (mentry.type != MFLOW_TRY) {
      continue;
    }
    auto& tentry = mentry.tentry;
    if (tentry->type == TRY_START) {
      always_assert(active_try == nullptr);
      active_try = &mentry;
      continue;
    }
    assert(tentry->type == TRY_END);
    auto try_end = &mentry;
    auto try_start = active_try;
    active_try = nullptr;

    always_assert_log(
        try_start != nullptr, "unopened try_end found: %s", SHOW(*try_end));
    always_assert_log(try_start->tentry->catch_start ==
                          try_end->tentry->catch_start,
                      "mismatched try start (%s) and end (%s)",
                      SHOW(*try_start),
                      SHOW(*try_end));
    auto insn_count = entry_to_addr.at(try_end) - entry_to_addr.at(try_start);
    if (insn_count == 0) {
      continue;
    }
    auto try_item = new DexTryItem(entry_to_addr.at(try_start), insn_count);
    for (auto mei = try_end->tentry->catch_start;
        mei != nullptr;
        mei = mei->centry->next) {
      try_item->m_catches.emplace_back(mei->centry->catch_type,
                                       entry_to_addr.at(mei));
    }
    tries.emplace_back(try_item);
  }
  always_assert_log(active_try == nullptr, "unclosed try_start found");

  std::sort(tries.begin(),
            tries.end(),
            [](const std::unique_ptr<DexTryItem>& a,
               const std::unique_ptr<DexTryItem>& b) {
              return a->m_start_addr < b->m_start_addr;
            });
  return true;
}

bool InstructionIterable::structural_equals(const InstructionIterable& other) {
  auto it1 = this->begin();
  auto it2 = other.begin();

  for (; it1 != this->end() && it2 != other.end(); it1++, it2++) {
    auto& mie1 = *it1;
    auto& mie2 = *it2;

    if (*mie1.insn != *mie2.insn) {
      return false;
    }
  }

  return it1 == this->end() && it2 == other.end();
}

IRInstruction* primary_instruction_of_move_result_pseudo(
    FatMethod::iterator it) {
  --it;
  always_assert(it->type == MFLOW_OPCODE &&
                it->insn->has_move_result_pseudo());
  return it->insn;
}

IRInstruction* move_result_pseudo_of(FatMethod::iterator it) {
  ++it;
  always_assert(it->type == MFLOW_OPCODE &&
                opcode::is_move_result_pseudo(it->insn->opcode()));
  return it->insn;
}
