/**
 * Copyright (c) 2017-present, Facebook, Inc.
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the
 * LICENSE file in the root directory of this source tree. An additional grant
 * of patent rights can be found in the PATENTS file in the same directory.
 */

#include "ReorderInterfaces.h"

#include <unordered_map>

#include "DexClass.h"
#include "DexUtil.h"
#include "Resolver.h"
#include "Walkers.h"

/**
 * This pass reorders Interface list for each class to improve the
 * linear walk of this list when there is an invocation of a function
 * defined in one of these Interfaces.
 *
 * This pass first computes the number of function invocations for each
 * Interface across the app and sorts the Interface list in descending order
 * of the number of invocations. An alphabetical sort is used for tie-breaks
 * to increase consistency across Classes.
 */
namespace {
typedef std::unordered_map<const DexType*, int> CallFrequencyMap;

/**
 * Helper class to implement the pass
 */
class ReorderInterfacesImpl {
 public:
  explicit ReorderInterfacesImpl(Scope& scope) : m_scope(scope) {}
  void run();

 private:
  // Map to store the number of method invokes for each Interface
  CallFrequencyMap m_call_frequency_map;

  // Pointer to the Scope object used for the pass
  Scope& m_scope;

  void compute_call_frequencies(IRInstruction* insn);
  void reorder_interfaces();
  void reorder_interfaces_for_class(DexClass* cls);
  std::deque<DexType*> sort_interfaces(
      const std::deque<DexType*>& unsorted_list);
};

/**
 * Run the pass by first computing call frequencies for each interface
 * then sorting the list of Interfaces for each Class.
 */
void ReorderInterfacesImpl::run() {
  // Check out each instruction and process if it is a function invoke
  walk_opcodes(m_scope,
               [](DexMethod*) { return true; },
               [this](DexMethod* /* unused */, IRInstruction* insn) {
                 compute_call_frequencies(insn);
               });

  // Now that we have the invoke frequencies for each Interface,
  // reorder the list of Interfaces for each Class.
  reorder_interfaces();
}

/**
 * Check whether the given instruction is a call to an Interface.
 * If so, increment the call frequency to that Interface.
 *
 * This method is used when we walk the opcodes in ReorderInterfacesImpl::run.
 */
void ReorderInterfacesImpl::compute_call_frequencies(IRInstruction* insn) {
  // Process only call instructions
  if (is_invoke(insn->opcode())) {
    auto callee = insn->get_method();
    auto def_callee = resolve_method(callee, opcode_to_search(insn));
    if (def_callee != nullptr) {
      callee = def_callee;
    }
    if (callee) {
      // Get the class this method is in. It may be an Interface or a Class
      const auto callee_cls_type = callee->get_class();
      const auto callee_cls = type_class(callee_cls_type);
      if (callee_cls) {
        // If we are calling into an Interface, count this call.
        if (is_interface(callee_cls)) {
          m_call_frequency_map[callee_cls_type]++;
        }
      }
    }
  }
}

/**
 * Sort the list of given Interfaces with respect to the number of incoming
 * calls and return the sorted list
 */
std::deque<DexType*> ReorderInterfacesImpl::sort_interfaces(
    const std::deque<DexType*>& unsorted_list) {
  std::deque<DexType*> sorted_list;
  // Create list of interfaces and store frequencies
  std::vector<std::pair<DexType*, int>> list_with_frequencies;
  for (auto interface : unsorted_list) {
    list_with_frequencies.emplace_back(interface,
                                       m_call_frequency_map[interface]);
  }

  // Sort the list with respect to number of calls for each Interface.
  std::sort(
      list_with_frequencies.begin(),
      list_with_frequencies.end(),
      [](const std::pair<DexType*, int>& a,
         const std::pair<DexType*, int>& b) -> bool {
        return ((b.second < a.second) ||
                ((b.second == a.second) && compare_dextypes(a.first, b.first)));
      });

  // Strip the frequencies
  for (auto interface_frequency : list_with_frequencies) {
    sorted_list.push_back(interface_frequency.first);
  }

  // Return the sorted list
  return sorted_list;
}

/**
 * Reorders Interface list for the given Class using the call frequencies
 */
void ReorderInterfacesImpl::reorder_interfaces_for_class(DexClass* cls) {
  const auto& cur_interface_list = cls->get_interfaces()->get_type_list();

  // If we have at most one interface implemented by this class,
  // no need to worry about sorting the interface list.
  if (cur_interface_list.size() <= 1) {
    return;
  }

  // Let's sort the Interface list
  auto updated_interface_list = sort_interfaces(cur_interface_list);

  DexTypeList* interfaces =
      DexTypeList::make_type_list(std::move(updated_interface_list));

  // Now that we have the sorted list of Interfaces, write it back.
  cls->set_interfaces(interfaces);
}

/**
 * Reorders Interface list for all Classes in the Scope
 */
void ReorderInterfacesImpl::reorder_interfaces() {
  for (auto cls : m_scope) {
    reorder_interfaces_for_class(cls);
  }
}
}

/**
 * Compute the number of function invocations for each Interface and
 * sort the list of Interfaces for each Class.
 */
void ReorderInterfacesPass::run_pass(DexStoresVector& stores,
                                     ConfigFiles& /* unused */,
                                     PassManager& /* unused */) {
  auto scope = build_class_scope(stores);

  ReorderInterfacesImpl impl(scope);
  impl.run();
}

static ReorderInterfacesPass ri_pass;
